"""Some type infos"""

from typing import AsyncIterable, Awaitable, Callable, Iterable, TypeVar, Union


T = TypeVar('T')    # pylint: disable=invalid-name

IterableFactorySync = Callable[[], Iterable[T]]
IterableFactoryAsync = Callable[[], Awaitable[Iterable[T]]]

AsyncIterableFactorySync = Callable[[], AsyncIterable[T]]
AsyncIterableFactoryAsync = Callable[[], Awaitable[AsyncIterable[T]]]

IterableAsync = Awaitable[Iterable[T]]
AsyncIterableAsync = Awaitable[AsyncIterable[T]]

GeneralIterable = Union[Iterable[T], IterableAsync[T], AsyncIterable[T], AsyncIterableAsync[T]]

IterableFactory = Union[IterableFactorySync[T], IterableFactoryAsync[T], AsyncIterableFactorySync[T], AsyncIterableFactoryAsync[T]]



"""Some general purpose utilities"""
import asyncio
from asyncio import (AbstractEventLoop, CancelledError, ensure_future,
                     iscoroutinefunction)
from functools import partial
from logging import getLogger
from typing import Any, Awaitable, Callable, Optional, TypeVar

from rx import Observable
from rx import create as rx_create
from rx.core.typing import Action, Observer, OnNext, Scheduler, TState
from rx.disposable import Disposable

T = TypeVar('T', covariant=True)    # pylint: disable=invalid-name

logger = getLogger(__name__)

_get_loop = getattr(asyncio, 'get_running_loop', None) or getattr(asyncio, '_get_running_loop')
assert callable(_get_loop)


async def _async_subscribe(generator: Callable[[OnNext, AbstractEventLoop], Awaitable], observer: Observer[T], event_loop: Optional[AbstractEventLoop]):
    #: sanity check
    assert observer
    #: access the event loop
    loop = event_loop or _get_loop()

    try:
        #: log this
        logger.debug('[%s]: Subscribing...')
        #: produce the value and dispatch
        await generator(observer.on_next, loop)
        loop.call_soon(observer.on_completed)
        #: log this
        logger.debug('[%s]: Completed')
    except CancelledError:
        loop.call_soon(observer.on_completed)
        #: log this
        logger.debug('[%s]: Canceled')
    except Exception as ex:  # pylint: disable=broad-except
        #: log this
        logger.exception('[%s]: Error')
        #: dispatch the error
        loop.call_soon(observer.on_error, ex)


def _create_action(delegate: Callable[[], Any]) -> Action:

    def _dispatch():
        delegate()

    return _dispatch


def _create_disposable(generator: Callable[[OnNext, AbstractEventLoop], Awaitable], observer: Observer[T], loop: Optional[AbstractEventLoop] = None) -> Disposable:
    return Disposable(_create_action(ensure_future(_async_subscribe(generator, observer, loop), loop=loop).cancel))


def _scheduled_action(generator: Callable[[OnNext, AbstractEventLoop], Awaitable], observer: Observer[T], loop: AbstractEventLoop, _scheduler: Scheduler, _state: Optional[TState]) -> Disposable:
    return _create_disposable(generator, observer, loop)


def _on_subscribe(generator: Callable[[OnNext, AbstractEventLoop], Awaitable], base_scheduler: Optional[Scheduler], observer: Observer[T], scheduler: Optional[Scheduler]):
    #: sanity check
    assert observer
    #: make sure to access the default scheduler
    sched: Optional[Scheduler] = base_scheduler or scheduler

    if sched is not None:
        return sched.schedule(partial(_scheduled_action, generator, observer, getattr(sched, '_loop', None)))

    logger.warning('No scheduler available')
    return _create_disposable(generator, observer)


def from_async(generator: Callable[[OnNext, AbstractEventLoop], Awaitable], scheduler: Optional[Scheduler] = None) -> Observable:
    """Creates an observable from a generic generator

    Args:
        generator: the async generator that produces the result
        scheduler: optionally the scheduler to run on

    Returns:
        observable that represents this operation
    """
    assert iscoroutinefunction(generator)

    return rx_create(partial(_on_subscribe, generator, scheduler))



# +-------------------------------------------------------------------+
# | Licensed Materials - Property of IBM                              |
# |                                                                   |
# | Hosting Appliance                                                 |
# |                                                                   |
# | Copyright IBM Corp. 2020 All Rights Reserved                      |
# |                                                                   |
# | US Government Users Restricted Rights - Use, duplication or       |
# | disclosure restricted by GSA ADP Schedule Contract with IBM Corp. |
# +-------------------------------------------------------------------+
"""Load the content of file"""
from json import load
from os.path import dirname, join
from typing import Any, Iterable

from pytest import mark
from rx_utils import from_iterable_factory

_ROOT = dirname(__file__)


def _load_file(name: str) -> Iterable[Any]:
    with open(name, 'r', encoding='utf-8') as file_pointer:
        yield load(file_pointer)


@mark.asyncio
async def test_async_files():
    file_ = from_iterable_factory(lambda: _load_file(join(_ROOT, 'data', 'generated.json')))

    result = await file_
    assert result



"""Creates an iterator from a factory of iterables"""
from asyncio import AbstractEventLoop, CancelledError
from functools import partial, reduce
from inspect import isawaitable
from itertools import takewhile
from logging import getLogger
from typing import Any, AsyncIterable, Iterable, Optional, Union, cast

import rx.operators as op
from rx import Observable, defer, empty, from_iterable
from rx.core.typing import Action, OnNext, Predicate, Scheduler
from rx.scheduler.eventloop import AsyncIOScheduler

from .async_utils import from_async
from .types import (AsyncIterableAsync, GeneralIterable, IterableAsync,
                    IterableFactory, T)

logger = getLogger(__name__)


def _noop():
    pass


def _is_iterable(obj: Any) -> bool:
    return callable(getattr(obj, '__iter__', None))


def _is_asynciterable(obj: Any) -> bool:
    return callable(getattr(obj, '__aiter__', None))


async def _async_noop():
    pass


async def _from_async_iterable(iterable: AsyncIterable[T], on_next: OnNext, loop: AbstractEventLoop):
    #: sanity check
    assert iterable
    assert callable(on_next)
    assert loop
    #: close callback
    done = getattr(iterable, 'aclose', _async_noop)
    try:
        async for value in iterable:
            loop.call_soon(on_next, value)
    finally:
        #: resource cleanup
        await done()


def _sync_generate(iterable: Iterable[T], on_next: OnNext, is_running: Predicate[T]):
    #: sanity checks
    assert callable(on_next)
    assert callable(is_running)

    done: Action = getattr(iterable, 'close', _noop)

    try:
        reduce(lambda _, item: on_next(item), takewhile(is_running, iter(iterable)), None)
    finally:
        done()


async def _async_generate(iterable: Iterable[T], on_next: OnNext, loop: AbstractEventLoop):
    #: sanity checks
    assert callable(on_next)
    assert loop

    #: cancelation flag
    is_running = True
    try:
        await loop.run_in_executor(None,
                                   _sync_generate,
                                   iterable,
                                   partial(loop.call_soon_threadsafe, on_next),
                                   lambda _: is_running
                                   )
    except CancelledError:
        is_running = False


async def _async_from_async(iterable: Union[AsyncIterable[T], Iterable[T]], on_next: OnNext, loop: AbstractEventLoop):
    #: check for a regular iterable
    if _is_iterable(iterable):
        await _async_generate(cast(Iterable[T], iterable), on_next, loop)
    elif _is_asynciterable(iterable):
        await _from_async_iterable(cast(AsyncIterable[T], iterable), on_next, loop)
    else:
        logger.warning('Object of type [%s] is neither Iterable nor AsyncIterable.', type(iterable))


def _from_async(scheduler: Optional[Scheduler], iterable: Union[AsyncIterableAsync[T], IterableAsync[T]]) -> Observable:
    #: sanity check
    assert isawaitable(iterable)

    async def _dispatch(on_next: OnNext, loop: AbstractEventLoop):
        await _async_from_async(await iterable, on_next, loop)

    return from_async(_dispatch, scheduler)


def _from_sync(scheduler: Optional[Scheduler], iterable: Union[AsyncIterable[T], Iterable[T]]) -> Observable:
    #: sanity check
    assert not isawaitable(iterable)

    #: check for a regular iterable
    if _is_iterable(iterable):
        return from_iterable(cast(Iterable[T], iterable), scheduler)

    async def _dispatch(on_next: OnNext, loop: AbstractEventLoop):
        await _from_async_iterable(cast(AsyncIterable[T], iterable), on_next, loop)

    #: check for an sync iterable
    if _is_asynciterable(iterable):
        return from_async(_dispatch, scheduler)

    logger.warning('Object of type [%s] is neither Iterable nor AsyncIterable.', type(iterable))

    return empty(scheduler)


async def _async_from_iterable_factory_on_async_scheduler(factory: IterableFactory[T], on_next: OnNext, loop: AbstractEventLoop):
    #: construct the resource
    iterable: GeneralIterable[T] = factory()

    if isawaitable(iterable):
        await _async_from_async(await cast(Union[AsyncIterableAsync[T], IterableAsync[T]], iterable), on_next, loop)
    else:
        await _async_from_async(cast(Union[AsyncIterable[T], Iterable[T]], iterable), on_next, loop)


def _from_iterable_factory_on_async_scheduler(factory: IterableFactory[T], scheduler: AsyncIOScheduler) -> Observable:

    async def _dispatch(on_next: OnNext, loop: AbstractEventLoop):
        await _async_from_iterable_factory_on_async_scheduler(factory, on_next, loop)

    return from_async(_dispatch, scheduler)


def _from_iterable_factory_on_scheduler(factory: IterableFactory[T], scheduler: Optional[Scheduler]) -> Observable:
    #: construct the resource
    iterable: GeneralIterable[T] = factory()
    done: Action = getattr(iterable, 'close', _noop)
    #: ensure cleanup
    return (_from_async if isawaitable(iterable) else _from_sync)(scheduler, iterable).pipe(op.finally_action(done))


def _from_iterable_factory(factory: IterableFactory[T], base_scheduler: Optional[Scheduler], scheduler: Optional[Scheduler]) -> Observable:
    #: the scheduler
    scheduler = scheduler or base_scheduler

    if isinstance(scheduler, AsyncIOScheduler):
        return _from_iterable_factory_on_async_scheduler(factory, scheduler)

    return _from_iterable_factory_on_scheduler(factory, scheduler)


def from_iterable_factory(factory: IterableFactory[T], scheduler: Optional[Scheduler] = None) -> Observable:
    """Constructs an observable for an sync or async iterable factory. Each subscription will create
    a new iterable via the factory, this allows e.g. to iterate multiple times over generators. If the object
    returned by the iterable factory is closeable (sync or async) then it will be closed when the
    observable gets unsubscribed. This is e.g. the case for generators.

    Args:
        factory: the iterable factory, may be sync or async
        scheduler: Optionally a scheduler. Defaults to None.

    Returns:
        the observable
    """
    #: quick sanity check
    assert callable(factory)
    #: dispatch
    return defer(partial(_from_iterable_factory, factory, scheduler))



# +-------------------------------------------------------------------+
# | Licensed Materials - Property of IBM                              |
# |                                                                   |
# | Hosting Appliance                                                 |
# |                                                                   |
# | Copyright IBM Corp. 2020 All Rights Reserved                      |
# |                                                                   |
# | US Government Users Restricted Rights - Use, duplication or       |
# | disclosure restricted by GSA ADP Schedule Contract with IBM Corp. |
# +-------------------------------------------------------------------+
"""Test the use of a generator to build an observable"""
from typing import Iterator
from contextlib import contextmanager
from rx_utils import from_iterable_factory
from rx.operators import take


def test_generator():

    entered = False
    released = False

    @contextmanager
    def my_context():
        nonlocal entered
        nonlocal released
        entered = True
        try:
            yield
        finally:
            released = True

    def my_generator() -> Iterator[int]:
        with my_context():
            yield 1
            yield 2
            yield 3

    obs = from_iterable_factory(my_generator).pipe(take(2))

    obs.subscribe()

    assert entered
    assert released



# +-------------------------------------------------------------------+
# | Licensed Materials - Property of IBM                              |
# |                                                                   |
# | Hosting Appliance                                                 |
# |                                                                   |
# | Copyright IBM Corp. 2020 All Rights Reserved                      |
# |                                                                   |
# | US Government Users Restricted Rights - Use, duplication or       |
# | disclosure restricted by GSA ADP Schedule Contract with IBM Corp. |
# +-------------------------------------------------------------------+

"""UtilsRxPY"""
from typing import Tuple

# version information
__version__ = '0.0.5'

from ._internal.from_iterable_factory import from_iterable_factory

# Some python nonsense
__all__: Tuple[str, ...] = (
    'from_iterable_factory',
)



# +-------------------------------------------------------------------+
# | Licensed Materials - Property of IBM                              |
# |                                                                   |
# | Hosting Appliance                                                 |
# |                                                                   |
# | Copyright IBM Corp. 2020 All Rights Reserved                      |
# |                                                                   |
# | US Government Users Restricted Rights - Use, duplication or       |
# | disclosure restricted by GSA ADP Schedule Contract with IBM Corp. |
# +-------------------------------------------------------------------+
from asyncio import AbstractEventLoop
from functools import partial
from glob import iglob
from logging import getLogger
from os.path import dirname, join
from typing import Iterator

from pytest import mark
import rx.operators as op
from rx_utils import from_iterable_factory

logger = getLogger(__name__)

_ROOT = dirname(__file__)


def _list_files() -> Iterator[str]:
    return iglob(join(_ROOT, '**', '*.*'), recursive=True)


@mark.asyncio
async def test_sync_iterable():
    obs_ = from_iterable_factory(_list_files).pipe(
        op.take(3),
        op.do_action(print)
    )

    await obs_



# +-------------------------------------------------------------------+
# | Licensed Materials - Property of IBM                              |
# |                                                                   |
# | Hosting Appliance                                                 |
# |                                                                   |
# | Copyright IBM Corp. 2020 All Rights Reserved                      |
# |                                                                   |
# | US Government Users Restricted Rights - Use, duplication or       |
# | disclosure restricted by GSA ADP Schedule Contract with IBM Corp. |
# +-------------------------------------------------------------------+

"""Exposes the inverse object mapping for sphinx"""
from typing import Sequence
from os.path import dirname, join, normpath

# version information
__version__ = '0.0.5'


def get_objects_inv_path() -> str:
    """Returns the pathname of the object.inv file

    Returns:
        the pathname
    """
    return normpath(join(dirname(__file__), 'objects.inv'))


# Some python nonsense
__all__: Sequence[str] = (
    'get_objects_inv_path',
)



# +-------------------------------------------------------------------+
# | Licensed Materials - Property of IBM                              |
# |                                                                   |
# | Hosting Appliance                                                 |
# |                                                                   |
# | Copyright IBM Corp. 2020 All Rights Reserved                      |
# |                                                                   |
# | US Government Users Restricted Rights - Use, duplication or       |
# | disclosure restricted by GSA ADP Schedule Contract with IBM Corp. |
# +-------------------------------------------------------------------+
from asyncio import sleep
from logging import getLogger
from typing import AsyncIterator

from pytest import mark
from rx.operators import map, take
from rx_utils import from_iterable_factory

logger = getLogger(__name__)


async def async_my_generator() -> AsyncIterator[int]:
    await sleep(0.1)
    yield 1
    await sleep(0.1)
    yield 2
    await sleep(0.1)
    yield 3


@mark.asyncio
async def test_async_iterable():
    obs_ = from_iterable_factory(async_my_generator).pipe(
        map(lambda x: x * x)
    )

    last = await obs_
    assert last == 9



# +-------------------------------------------------------------------+
# | Licensed Materials - Property of IBM                              |
# |                                                                   |
# | Hosting Appliance                                                 |
# |                                                                   |
# | Copyright IBM Corp. 2020 All Rights Reserved                      |
# |                                                                   |
# | US Government Users Restricted Rights - Use, duplication or       |
# | disclosure restricted by GSA ADP Schedule Contract with IBM Corp. |
# +-------------------------------------------------------------------+

""" Setup file """

from pathlib import Path
from typing import Sequence
from shutil import copy

from setuptools import setup, find_packages


def copy_inv():
    try:
        root = Path(__file__).parent
        copy(root / 'build' / 'docs' / 'objects.inv', root / 'rx_utils' / 'docs')
    except BaseException:
        pass

def read(fname: str) -> str:
    return open(Path(__file__).parent / fname).read()


def read_requirements(filename: str) -> Sequence[str]:
    return read(filename).splitlines()


settings = dict(
    name='UtilsRxPY',
    packages=find_packages(exclude=["test"]),
    version='0.0.5',
    author='Dr. Carsten Leue',
    author_email='carsten.leue@gmx.net',
    description=('This is an awesome project!'),
    license='MIT',
    keywords='UtilsRxPY',
    url='https://github.com/Carsten-Leue/UtilsRxPY',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    include_package_data=True,
    python_requires='>=3.6',
    install_requires=read_requirements('requirements.txt'),
    tests_require=read_requirements('test-requirements.txt'),
    classifiers=[
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',        
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Typing :: Typed'
    ],
    zip_safe=False
)

copy_inv()
setup(**settings)
